@@ -63,9 +63,11 @@ type action =
   | Accept
   | Clear_screen
   | Prev_search
+  | Next_search
   | Cancel_search
   | Break
   | Suspend
+  | Edit_with_external_editor
 
 let doc_of_action = function
   | Edit action -> LTerm_edit.doc_of_action action
@@ -81,9 +83,11 @@ let doc_of_action = function
   | Accept -> "accept the current input."
   | Clear_screen -> "clear the screen."
   | Prev_search -> "search backward in the history."
+  | Next_search -> "search forward in the history."
   | Cancel_search -> "cancel search mode."
   | Break -> "cancel edition."
   | Suspend -> "suspend edition."
+  | Edit_with_external_editor -> "edit input with external editor command."
 
 let actions = [
   Interrupt_or_delete_next_char, "interrupt-or-delete-next-char";
@@ -98,9 +102,11 @@ let actions = [
   Accept, "accept";
   Clear_screen, "clear-screen";
   Prev_search, "prev-search";
+  Next_search, "next-search";
   Cancel_search, "cancel-search";
   Break, "break";
   Suspend, "suspend";
+  Edit_with_external_editor, "edit-with-external-editor";
 ]
 
 let actions_to_names = Array.of_list (List.sort (fun (a1, n1) (a2, n2) -> Pervasives.compare a1 a2) actions)
@@ -164,6 +170,7 @@ let () =
   bind [{ control = true; meta = false; shift = false; code = Char(UChar.of_char 'm') }] [Accept];
   bind [{ control = true; meta = false; shift = false; code = Char(UChar.of_char 'l') }] [Clear_screen];
   bind [{ control = true; meta = false; shift = false; code = Char(UChar.of_char 'r') }] [Prev_search];
+  bind [{ control = true; meta = false; shift = false; code = Char(UChar.of_char 's') }] [Next_search];
   bind [{ control = true; meta = false; shift = false; code = Char(UChar.of_char 'd') }] [Interrupt_or_delete_next_char];
   bind [{ control = false; meta = true; shift = false; code = Char(UChar.of_char 'p') }] [History_prev];
   bind [{ control = false; meta = true; shift = false; code = Char(UChar.of_char 'n') }] [History_next];
@@ -174,7 +181,10 @@ let () =
   bind [{ control = false; meta = true; shift = false; code = Tab }] [Complete_bar];
   bind [{ control = false; meta = true; shift = false; code = Down }] [Complete_bar];
   bind [{ control = false; meta = true; shift = false; code = Enter }] [Edit (LTerm_edit.Zed Zed_edit.Newline)];
-  bind [{ control = false; meta = false; shift = false; code = Escape }] [Cancel_search]
+  bind [{ control = false; meta = false; shift = false; code = Escape }] [Cancel_search];
+  bind [{ control = true; meta = false; shift = false; code = Char(UChar.of_char 'x') }
+       ;{ control = true; meta = false; shift = false; code = Char(UChar.of_char 'e') }]
+    [Edit_with_external_editor]
 
 (* +-----------------------------------------------------------------+
    | The read-line engine                                            |
@@ -217,6 +227,14 @@ let no_completion = {
   count = 0;
 }
 
+type direction = Forward | Backward
+
+type search_status =
+  { before : Zed_utf8.t list
+  ; after  : Zed_utf8.t list
+  ; match_ : (Zed_utf8.t * int) option
+  }
+
 class virtual ['a] engine ?(history = []) ?(clipboard = LTerm_edit.clipboard) ?(macro = macro) () =
   let edit : unit Zed_edit.t = Zed_edit.create ~clipboard () in
   let context = Zed_edit.context edit (Zed_edit.new_cursor edit) in
@@ -300,41 +318,86 @@ object(self)
   (* The event which search for the string in the history. *)
   val mutable search_event = E.never
 
-  (* The result of the search. If the search was successful it
-     contains the matched history entry, the position of the substring
-     in this entry and the rest of the history. *)
-  val mutable search_result = None
+  val mutable search_status = None
 
   initializer
-    search_event <- E.map (fun _ -> search_result <- None; self#search) (E.when_ (S.map (fun mode -> mode = Search) mode) (Zed_edit.changes edit))
-
-  method private search =
-    let input = Zed_rope.to_string (Zed_edit.text edit) in
-    let rec loop = function
-      | [] ->
-          search_result <- None;
-          set_message (Some(LTerm_text.of_string "Reverse search: not found"))
-      | entry :: rest ->
-          match search_string entry input with
-            | Some pos -> begin
-                match search_result with
-                  | Some(entry', _, _) when entry = entry' ->
-                      loop rest
-                  | _ ->
-                      search_result <- Some(entry, pos, rest);
-                      let txt = LTerm_text.of_string entry in
-                      for i = pos to pos + Zed_rope.length (Zed_edit.text edit) - 1 do
-                        let ch, style = txt.(i) in
-                        txt.(i) <- (ch, { style with underline = Some true })
-                      done;
-                      set_message (Some(Array.append (LTerm_text.of_string "Reverse search: ") txt))
-              end
-            | None ->
-                loop rest
+    let reset_search _ =
+      search_status <- None;
+      self#search Backward
     in
-    match search_result with
-      | Some(entry, pos, rest) -> loop rest
-      | None -> loop (fst (S.value history))
+    search_event <-
+      E.map reset_search
+        (E.when_ (S.map (fun mode -> mode = Search) mode)
+           (Zed_edit.changes edit))
+
+  method private search direction =
+    let do_search direction =
+      let set_status other_entries entries match_ =
+        let before, after =
+          match direction with
+          | Backward -> (other_entries, entries)
+          | Forward  -> (entries, other_entries)
+        in
+        search_status <- Some { before; after; match_ }
+      in
+      let input = Zed_rope.to_string (Zed_edit.text edit) in
+      let rec loop other_entries entries =
+        match entries with
+        | [] ->
+          set_status other_entries entries None;
+          set_message (Some(LTerm_text.of_string "Reverse search: not found"))
+        | entry :: rest ->
+          match search_string entry input with
+          | Some pos -> begin
+              match search_status with
+              | Some { match_ = Some (entry', _) } when entry = entry' ->
+                loop (entry :: other_entries) rest
+              | _ ->
+                set_status other_entries rest (Some (entry, pos));
+                let txt = LTerm_text.of_string entry in
+                for i = pos to pos + Zed_rope.length (Zed_edit.text edit) - 1 do
+                  let ch, style = txt.(i) in
+                  txt.(i) <- (ch, { style with underline = Some true })
+                done;
+                set_message
+                  (Some (Array.append (LTerm_text.of_string "Reverse search: ") txt))
+            end
+          | None ->
+            loop (entry :: other_entries) rest
+      in
+      match search_status with
+      | None ->
+        let hist = fst (S.value history) in
+        loop []
+          (match direction with
+           | Backward -> hist
+           | Forward  -> List.rev hist)
+      | Some { before; after; match_ } ->
+        let other_entries, entries =
+          match direction with
+          | Backward -> (before, after)
+          | Forward  -> (after, before)
+        in
+        let other_entries =
+          match match_ with
+          | None -> other_entries
+          | Some (entry, _) -> entry :: other_entries
+        in
+        loop other_entries entries
+    in
+    match S.value mode with
+    | Search -> do_search direction
+    | Edition ->
+      let text = Zed_edit.text edit in
+      Zed_edit.goto context 0;
+      Zed_edit.remove context (Zed_rope.length text);
+      let prev, next = S.value history in
+      set_history (Zed_rope.to_string text :: (List.rev_append next prev), []);
+      search_status <- None;
+      set_mode Search;
+      do_search direction
+    | _ ->
+      ()
 
   method insert ch =
     Zed_edit.insert context (Zed_rope.singleton ch)
@@ -345,13 +408,13 @@ object(self)
       | (Complete | Complete_bar | Accept) when S.value mode = Search -> begin
           set_mode Edition;
           set_message None;
-          match search_result with
-            | Some(entry, pos, rest) ->
-                search_result <- None;
+          match search_status with
+            | Some { match_ = Some (entry, pos) } ->
+                search_status <- None;
                 Zed_edit.goto context 0;
                 Zed_edit.remove context (Zed_rope.length (Zed_edit.text edit));
                 Zed_edit.insert context (Zed_rope.of_string entry)
-            | None ->
+            | Some { match_ = None } | None ->
                 ()
         end
 
@@ -422,22 +485,8 @@ object(self)
                 Zed_edit.insert context (Zed_rope.of_string line)
         end
 
-      | Prev_search -> begin
-          match S.value mode with
-            | Search ->
-                self#search
-            | Edition ->
-                let text = Zed_edit.text edit in
-                Zed_edit.goto context 0;
-                Zed_edit.remove context (Zed_rope.length text);
-                let prev, next = S.value history in
-                set_history (Zed_rope.to_string text :: (List.rev_append next prev), []);
-                search_result <- None;
-                set_mode Search;
-                self#search
-            | _ ->
-                ()
-        end
+      | Prev_search -> self#search Backward
+      | Next_search -> self#search Forward
 
       | Cancel_search ->
           if S.value mode = Search then begin
@@ -580,7 +629,7 @@ class read_password () = object(self)
   method show_box = false
 
   method send_action = function
-    | Prev_search -> ()
+    | Prev_search | Next_search -> ()
     | action -> super#send_action action
 end
 
@@ -1060,6 +1109,14 @@ object(self)
       | _ ->
           self#loop
 
+  method create_temporary_file_for_external_editor =
+    Filename.temp_file "lambda-term" ".txt"
+
+  method external_editor =
+    try
+      Sys.getenv "EDITOR"
+    with Not_found -> "vi"
+
   method private exec = function
     | Accept :: _ when S.value self#mode = Edition ->
         Zed_macro.add self#macro Accept;
@@ -1101,6 +1158,59 @@ object(self)
           (if is_visible then self#show else return ()) >>= fun () ->
           self#exec actions
         end
+    | Edit_with_external_editor :: actions -> begin
+        let is_visible = visible in
+        self#hide >>= fun () ->
+        LTerm.flush term >>= fun () ->
+        begin
+          match mode with
+          | Some mode ->
+            LTerm.leave_raw_mode term mode
+          | None ->
+            return ()
+        end >>= fun () ->
+        let temp_fn = self#create_temporary_file_for_external_editor in
+        let input = Zed_rope.to_string (Zed_edit.text self#edit) in
+        Lwt_io.with_file ~mode:Output temp_fn (fun oc -> Lwt_io.write_line oc input)
+        >>= fun () ->
+        let editor = self#external_editor in
+        Printf.ksprintf Lwt_unix.system "%s %s" editor (Filename.quote temp_fn)
+        >>= fun status ->
+        (if status <> WEXITED 0 then
+           Lwt_io.eprintf "`%s %s' exited with status %d\n"
+             editor temp_fn
+             (match status with
+              | WEXITED n -> n
+              | _         -> 255)
+         else
+           Lwt.try_bind
+             (fun () -> Lwt_io.with_file ~mode:Input temp_fn Lwt_io.read)
+             (fun s  ->
+                let s = Zed_utf8.rstrip s in
+                Zed_edit.goto_bot self#context;
+                Zed_edit.replace self#context (Zed_rope.length (Zed_edit.text self#edit))
+                  (Zed_rope.of_string s);
+                Lwt.return ())
+             (function
+               | Unix.Unix_error (err, _, _) ->
+                 Lwt_io.eprintf "%s: %s\n" temp_fn (Unix.error_message err)
+               | exn -> Lwt.fail exn)
+        )
+        >>= fun () ->
+        begin
+          match LTerm.is_a_tty term with
+          | true ->
+            LTerm.enter_raw_mode term >>= fun m ->
+            mode <- Some m;
+            return ()
+          | false ->
+            return ()
+        end
+        >>= fun () ->
+        (if is_visible then self#show else return ())
+        >>= fun () ->
+        self#exec actions
+      end
     | action :: actions ->
         self#send_action action;
         self#exec actions
